from typing import Optional, Type, TYPE_CHECKING
from fastapi import APIRouter, HTTPException, status, Response, Request
from fastapi.responses import RedirectResponse
from datetime import datetime
from common.logger import get_logger
from server.settings import settings, OpenIDIssuer, SecretProvider

if TYPE_CHECKING:
    from server.auth.base import AuthBase
    from server.secret.base import SecretBase

logger = get_logger(__name__)

router = APIRouter(prefix="/auth")

def get_provider() -> Type["AuthBase"]:
    """determine provider from settings and return the appropriate class"""
    match settings.openid_issuer:
        case OpenIDIssuer.microsoft:
            from server.auth.microsoft import MicrosoftAuth
            return MicrosoftAuth()
        case OpenIDIssuer.keycloak:
            from server.auth.keycloak import KeycloakAuth
            return KeycloakAuth()
        case _:
            raise ValueError(f"Unsupported OpenID issuer: {settings.openid_issuer}")

def get_secret(name: str) -> Type["SecretBase"]:
    """determine secret provider from settings and return the appropriate class"""
    match settings.secret_provider:
        case SecretProvider.aws_secrets_manager:
            from server.secret.aws_secrets_manager import AWSSecretsManagerSecret
            return AWSSecretsManagerSecret(name)
        case _:
            raise ValueError(f"Unsupported Secret Provider: {settings.secret_provider}")

def set_auth_cookie(response: Response, key: str, value: str, secure: bool = True):
    """Helper function to set authentication cookies with consistent settings"""
    logger.info(f"Setting cookie: {key}")
    response.set_cookie(
        key=key,
        value=value,
        httponly=secure,
        secure=secure,
        samesite="lax",
        max_age=settings.max_age,
    )

def set_auth_cookies(response: Response,
                     user: str,
                     token: str,
                     refresh: str,
                     expires: int)->None:
    """set the authentication cookies"""
    logger.info("Setting authentication cookies")
    set_auth_cookie(response, "id_token", token)
    set_auth_cookie(response, "refresh_token", refresh)
    set_auth_cookie(response, "user", user)
    set_auth_cookie(response, "openid_expires", expires)
    logger.info("All cookies set successfully")

def delete_auth_cookies(response: Response)->None:
    """delete the authentication cookies"""
    logger.info("Deleting authentication cookies")
    response.delete_cookie("id_token")
    response.delete_cookie("refresh_token")
    response.delete_cookie("user")
    response.delete_cookie("openid_expires")
    logger.info("All cookies deleted successfully")




@router.get("/login")
async def login():
    """Log in via the OpenID provider"""
    provider = get_provider()
    return RedirectResponse(url=provider.get_auth_url())

@router.get("/callback")
async def auth_callback(code: Optional[str] = None,
                        error: Optional[str] = None):
    """Handle the callback from the OpenID provider"""
    logger.info("Processing auth callback...")
    if error:
        logger.error(f"Auth error: {error}")
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=f"Authentication failed: {error}",
        )
    if not code:
        logger.error("No authorization code provided")
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="No authorization code provided",
        )
    provider = get_provider()

    result = provider.exchange_code_for_token(code)
    if "error" in result:
        error_description = result.get(
            "error_description", "No error description available"
        )
        logger.error(
            f"Token acquisition failed - Error: {result['error']}, Description: {error_description}"
        )
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=f"Token acquisition failed: {result['error']} - {error_description}",
        )
    user = result["id_token_claims"]["preferred_username"]
    token = result["id_token"]
    refresh = result["refresh_token"]
    expires = result["id_token_claims"]["exp"]

    secret = get_secret(user)
    secret_identifier = "=".join(secret.secret_identity_pair)
    response = RedirectResponse(url=f"/?{secret_identifier}", status_code=status.HTTP_303_SEE_OTHER)
    secret.set_token_value(token)
    set_auth_cookies(response, user, token, refresh, expires)
    logger.info("Redirecting to home page")
    return response


@router.get("/logout")
async def logout(request: Request,
                 response: Response):
    """Log out the user"""
    user = request.cookies.get("user")
    logger.info(f"Logging out user {user}")
    secret = get_secret(user)
    secret.delete_token_value()
    response = RedirectResponse(
        url="/", status_code=status.HTTP_303_SEE_OTHER
    )
    delete_auth_cookies(response)
    return response





from typing import Optional
from pathlib import Path
from pydantic import Field
from enum import Enum
from common.settings import CommonSettings, SecretProvider

class Settings(CommonSettings):

    # Redirect base
    server_url: Optional[str] = Field(None, description="The base URL for the server. If set, the client will open a login window when the token is expired.")

    # Secret settings
    secret_provider: SecretProvider = Field(..., description="The provider where the secret is stored")
    secret_identifier: str = Field(..., description="The identifier for the secret")
    secret_cache_path: Optional[Path] = Field(Path.home() / ".locknessie" / "token", description="The path to where the token cache is stored")

settings = Settings()


from keycloak import KeycloakOpenID
from common.logger import get_logger
from server.settings import settings

logger = get_logger(__name__)


class KeycloakAuth:

    def __init__(self):

        self.keycloak_app = KeycloakOpenID(
            server_url=settings.openid_url,
            client_id=settings.openid_client_id,
            realm_name=settings.openid_realm,
            client_secret_key=settings.openid_client_secret,
        )

    def get_auth_url(self) -> str:
        return self.keycloak_app.auth_url(
            scope="User.Read",
            redirect_url=self.redirect_uri
        )

    def exchange_code_for_token(self, code: str) -> str:
        self.logger.info("Exchanging code for tokens")
        return self.keycloak_app.token(
            grant_type="authorization_code",
            code=code,
            redirect_uri=self.redirect_uri,
        )



from typing import TYPE_CHECKING, Optional
from common.logger import get_logger
from server.settings import settings

if TYPE_CHECKING:
    from logging import Logger

class AuthBase:
    logger: "Logger"
    redirect_uri: str

    def __init__(self, redirect_uri: Optional[str]=None):
        self.logger = get_logger(self.__class__.__name__)
        self.redirect_uri = redirect_uri or f"{settings.redirect_base}/auth/callback"



from typing import TYPE_CHECKING, Optional
from pyiceberg import load_catalog
from client.secret.base import get_secret_provider

if TYPE_CHECKING:
    from pyiceberg import Catalog

def load_catalog(name: Optional[str] = None, **properties: Optional[str]) -> Catalog:
    provider = get_secret_provider()
    return load_catalog(name, **properties, token=provider.token)


from typing import Optional, Type
from abc import ABC, abstractmethod
from client.settings import settings, ExpiredAction, SecretProvider

def get_secret_provider() -> Type["SecretBase"]:
    match settings.secret_provider:
        case SecretProvider.aws_secrets_manager:
            from client.secret.aws_secrets_manager import AwsSecretsManager
            return AwsSecretsManager()
        case _:
            raise ValueError(f"Unknown secret provider: {settings.secret_provider}")

class ExpiredTokenError(Exception):
    """Exception raised when a token is expired"""

class SecretBase(ABC):
    """Base class for secret stores"""
    secret_identifier:str

    def __init__(self, secret_identifier: Optional[str] = None):
        self.secret_identifier = secret_identifier or settings.secret_identifier
        assert self.secret_identifier, "`secret_identifier` is required as either a command line argument or environment variable"

    @property
    def token(self) -> Optional[str]:
        if settings.secret_cache_path.exists():
            return settings.secret_cache_path.read_text()
        else:
            token = self.retrieve_token()
            settings.secret_cache_path.parent.mkdir(parents=True, exist_ok=True)
            settings.secret_cache_path.write_text(token, encoding="utf-8")
            return token

    @abstractmethod
    def retrieve_token(self) -> Optional[str]:
        """get the token from the secret store"""




from enum import Enum
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing import Optional

class SecretProvider(str, Enum):
    aws_secrets_manager = "aws_secrets_manager"

class CommonSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="locknessie_", case_sensitive=False)

    deleted_string: str = Field("DELETED", description="The string to use when the secret is deleted")


from enum import Enum
from pathlib import Path
from pydantic import Field
from typing import Optional
from common.settings import CommonSettings, SecretProvider

class OpenIDIssuer(str, Enum):
    microsoft = "microsoft"
    keycloak = "keycloak"

class Settings(CommonSettings):

    environment: str = Field(..., description="'production' for released code, 'development' for local development")

    # Redirect base
    redirect_base: str = Field(..., description="The base URL for redirects")

    # OpenID settings
    openid_issuer: OpenIDIssuer = Field(..., description="The issuer of the OpenID client")
    openid_client_id: str = Field(..., description="The client ID of the OpenID client")
    openid_client_secret: str = Field(..., description="The client secret of the OpenID client")
    openid_tenant: Optional[str] = Field(None, description="The tenant of the OpenID client")
    openid_realm: Optional[str] = Field(None, description="The realm of the OpenID client")
    openid_url: Optional[str] = Field(None, description="The URL of the OpenID provider")

    # Secret settings
    secret_provider: SecretProvider = Field(..., description="The provider where the secret is stored")

    # Cookie settings
    max_age: int = Field(default=(60 * 60 * 24 * 365), description="The maximum age of the cookie in seconds, default is 1 year")

    # Template settings
    templates_dir_path: str = Field(default=str(Path(__file__).parent / "templates"), description="The path to locate the home.html template for rendering the home page.")


settings = Settings()


from typing import Optional
import boto3
import webbrowser
from client.settings import settings, ExpiredAction
from client.secret.base import SecretBase, ExpiredTokenError

class AWSSecretsManagerSecret(SecretBase):
    client: "boto3.client"

    def __init__(self,
                 secret_identifier: Optional[str] = None):
        super().__init__(secret_identifier)
        self.client = boto3.client("secretsmanager")


    def retrieve_token(self) -> Optional[str]:
        secret = self.client.get_secret_value(SecretId=self.secret_identifier)
        value = secret["SecretString"]
        if not (value == settings.deleted_string):
            return value
        if settings.server_url:
            webbrowser.open(f"{settings.server_url}/auth/login")
        raise ExpiredTokenError("Token expired, please login to the auth server to continue")


import click
import uvicorn
import importlib.metadata as metadata
from typing import Optional
from datetime import datetime
from fastapi import FastAPI, Request
from fastapi.responses import RedirectResponse, HTMLResponse
from fastapi.templating import Jinja2Templates
from fastapi.staticfiles import StaticFiles
from pathlib import Path
import humanize
from common.logger import get_logger
from server.settings import settings
from server.main import router

app = FastAPI(title="Lock-Nessie Auth Server", version=metadata.version("locknessie"))
app.include_router(router)

# Get the directory containing the current file
templates = Jinja2Templates(directory=settings.templates_dir_path)

# Mount static files
app.mount("/static", StaticFiles(directory=str(Path(__file__).parent / "static")), name="static")

logger = get_logger(__name__)

@app.get("/")
def home(request: Request,
         aws_secret_arn: Optional[str] = None):
    """Home page"""
    user = request.cookies.get("user", None)
    expires = request.cookies.get("openid_expires", None)
    is_logged_in = bool(user and expires)

    if is_logged_in:
        expires_dt = datetime.fromtimestamp(int(expires))
        now = datetime.now()
        time_until_expiry = humanize.naturaldelta(expires_dt - now)
        logger.info(f"User {user} found, returning home page")
    else:
        logger.info("No user found, showing login page")
        user = None
        expires_dt = None
        time_until_expiry = None

    return templates.TemplateResponse(
        "home.html",
        {
            "request": request,
            "user": user,
            "is_logged_in": is_logged_in,
            "expires": expires_dt,
            "time_until_expiry": time_until_expiry,
            "aws_secret_arn": aws_secret_arn
        }
    )


@click.group()
def cli():
    """Lock-Nessie CLI tool"""
    pass

@cli.command()
@click.option("--host", default="0.0.0.0", help="Host to bind the server to")
@click.option("--port", default=8080, help="Port to bind the server to")
@click.option("--reload", is_flag=True, help="Enable auto-reload on code changes")
def server(host: str, port: int, reload: bool):
    """Start the Lock-Nessie authentication server"""
    uvicorn.run(
        "server.cli:app",
        host=host,
        port=port,
        reload=reload,
    )

if __name__ == "__main__":
    cli()


from server.secret.base import SecretBase
import boto3

class AWSSecretsManagerSecret(SecretBase):

    def __init__(self, user: str):
        super().__init__(user)
        self.client = boto3.client("secretsmanager")
        self.secret_arn = self.create_secret()

    def set_token_value(self, token: str) -> None:
        """set the token value in the secret store"""
        self.create_secret()
        self.logger.info(f"Setting token value for {self.user}")
        result = self.client.put_secret_value(
            SecretId=self.secret_name,
            SecretString=token,
        )
        if result["ResponseMetadata"]["HTTPStatusCode"] != 200:
            raise Exception(f"Failed to set token value for {self.user}: {result}")
        self.logger.info(f"Token value set for {self.user}")

    def delete_token_value(self):
        """delete the token value from the secret store"""
        self.logger.info(f"Setting empty token value for {self.user}")
        self.create_secret()
        self.set_token_value(self.deleted_string)


    def create_secret(self) -> str:
        """create the empty secret if it doesn't already exist, and return the secret arn"""
        self.logger.info(f"Creating secret for {self.user}")
        try:
            secret = self.client.create_secret(
                Name=self.secret_name,
                SecretString=self.deleted_string,
            )
        except self.client.exceptions.ResourceExistsException:
            self.logger.info(f"Secret already exists for {self.user}, skipping creation")
            secret = self.client.describe_secret(SecretId=self.secret_name)
        except Exception as e:
            self.logger.error(f"Error creating secret for {self.user}: {e}")
            raise e
        return secret["ARN"]

    @property
    def secret_identity_pair(self) -> tuple[str, str]:
        """get the secret arn"""
        return "aws_secret_arn", self.secret_arn


from abc import ABC, abstractmethod
import re
from common.logger import get_logger
from server.settings import settings

class SecretBase(ABC):
    user: str
    secret_name: str
    logger: "Logger"
    deleted_string:str = "DELETED"

    def __init__(self, user: str):
        self.logger = get_logger(self.__class__.__name__)
        self.user = user
        self.secret_name = self.build_secret_name(user)

    def build_secret_name(self, user: str):
        """build the secret name from the user"""
        safe_user = re.sub(r"[^a-zA-Z0-9]", "-", user).strip("-")
        return f"{settings.openid_client_id}-{safe_user}"

    @abstractmethod
    def set_token_value(self, secret: str):
        """set the token value in the secret store"""

    @abstractmethod
    def delete_token_value(self):
        """delete the token value from the secret store"""


from msal import ConfidentialClientApplication
from common.logger import get_logger
from server.settings import settings
from server.auth.base import AuthBase

logger = get_logger(__name__)


class MicrosoftAuth(AuthBase):

    def __init__(self):
        super().__init__()
        self.authority = f"https://login.microsoftonline.com/{settings.openid_tenant}"
        logger.info(f"Initializing MSAL with authority: {self.authority}")
        logger.info(f"Client ID length: {len(settings.openid_client_id)}")

        self.msal_app = ConfidentialClientApplication(
            client_id=settings.openid_client_id,
            client_credential=settings.openid_client_secret,
            authority=self.authority,
        )

    def get_auth_url(self) -> str:
        return self.msal_app.get_authorization_request_url(
            scopes=["User.Read"], redirect_uri=self.redirect_uri, response_type="code"
        )

    def exchange_code_for_token(self, code: str) -> str:
        self.logger.info("Exchanging code for tokens")
        return self.msal_app.acquire_token_by_authorization_code(
            code=code,
            scopes=["User.Read"],
            redirect_uri=self.redirect_uri,
        )
